package org.fortiss.variability.analysis;

import static com.microsoft.z3.Status.UNSATISFIABLE;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.variability.VariabilityActivator.getDefault;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.tooling.kernel.model.INamedElement;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Context;
import com.microsoft.z3.Expr;
import com.microsoft.z3.FuncDecl;
import com.microsoft.z3.Model;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Status;

/**
 * Base class that implements the translation of model product-lines to SMT in order to perform
 * product-line analysis via constraint lifting.
 * 
 * @author bayha
 */
public abstract class GenericProductLineAnalysis {

	protected final IProductLineTranslation translation;

	/** List of all translated correctness constraints. */
	protected List<BoolExpr> rawConstraints = new ArrayList<BoolExpr>();

	protected List<IProductLineConstraint> plConstraints = new ArrayList<IProductLineConstraint>();

	protected Map<BoolExpr, IProductLineConstraint> constraintTracker2Constraint =
			new HashMap<BoolExpr, IProductLineConstraint>();

	protected final EObject model;

	protected Context ctx;

	/**
	 * Constructor.
	 */
	protected GenericProductLineAnalysis(EObject model, IProductLineTranslation translation) {
		this.model = model;

		this.translation = translation;
		ctx = translation.getContext();
	}

	/**
	 * Performs a product-line analysis for the given model.
	 * 
	 * @return 'true' if all constraints are fulfilled for all variants of the product-line
	 *         analysis. 'false' otherwise.
	 */
	public List<ProductLineConstraintViolation> doCheck() {
		final long start = System.currentTimeMillis();

		addConstriants();

		setTranslatedClassesAndReferences();

		translation.translateModel(model);

		constraintTracker2Constraint = translation.translateConstraints(plConstraints);

		// addRawConstraints();

		// Global.setParameter("parallel.enable", "true");

		Solver solver = translation.createSolver(tracToExpr);

		String modelName = ((INamedElement)model).getName();
		String path = "C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "in";
		try {
			PrintWriter printWriter = new PrintWriter(path);
			printWriter.println(solver);
			printWriter.println("(check-sat)");
			// printWriter.println("(get-model)");
			printWriter.close();
		} catch(FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		long startSmt = System.currentTimeMillis();

		Status result = solver.check();

		long end = System.currentTimeMillis();

		System.out.println("Runtime " + modelName + ": " + (end - start) + " total, " +
				(startSmt - start) + " translation, " + (end - startSmt) + " SMT");

		List<ProductLineConstraintViolation> ret = new ArrayList<ProductLineConstraintViolation>();

		if(result.equals(UNSATISFIABLE)) {
			// No constraint was violated

			// BoolExpr[] unsatCore = solver.getUnsatCore();
			//
			// PrintWriter printWriter;
			//
			// try {
			// printWriter = new PrintWriter(
			// "C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "unsat");
			//
			// for(BoolExpr b : unsatCore) {
			// BoolExpr coreExpr = tracToExpr.get(b);
			//
			// printWriter.println(coreExpr);
			// }
			//
			// printWriter.close();
			// } catch(FileNotFoundException e) {
			// // TODO Auto-generated catch block
			// e.printStackTrace();
			// }

			// System.out.println(deepToString(unsatCore));
		} else {

			Model z3Model = solver.getModel();

			for(BoolExpr constTrack : constraintTracker2Constraint.keySet()) {
				Expr constInterp = z3Model.getConstInterp(constTrack);

				if(constInterp instanceof BoolExpr) {
					if(constInterp.equals(ctx.mkTrue())) {
						IProductLineConstraint failedConstraint =
								constraintTracker2Constraint.get(constTrack);

						List<EObject> violatingObjects = new ArrayList<EObject>();
						for(Expr varTrack : failedConstraint.getVariableTracker()) {
							Expr interpr = z3Model.getConstInterp(varTrack);
							EObject eObject = translation.getEObjectForExpression(interpr);

							violatingObjects.add(eObject);
						}

						ret.add(new ProductLineConstraintViolation(failedConstraint,
								violatingObjects, null));
					}
				}

			}

			try {
				PrintWriter printWriter = new PrintWriter(
						"C:\\Users\\bayha\\tmp\\z3" + modelName + (start % 159739431) + "out");
				printWriter.println(z3Model);
				printWriter.close();
			} catch(FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		translation.getContext().close();

		return ret;
	}

	/**
	 * 
	 */
	private void setTranslatedClassesAndReferences() {
		Set<EClass> classes = new HashSet<EClass>();
		Set<EReference> references = new HashSet<EReference>();
		Set<EAttribute> attributes = new HashSet<EAttribute>();

		for(IProductLineConstraint constraint : plConstraints) {
			classes.addAll(constraint.getTranslatedClasses());
			references.addAll(constraint.getTranslatedReferences());
			attributes.addAll(constraint.getTranslatedAttributes());
		}

		translation.setTranslatedEClasses(classes);
		translation.setTranslatedEReferences(references);
		translation.setTranslatedEAttributes(attributes);

	}

	/** Mapping from tracking variables to z3 expression (used for counterexample extraction). */
	protected Map<BoolExpr, BoolExpr> tracToExpr = new HashMap<BoolExpr, BoolExpr>();

	// /** gets the Z3 solver to be used for analysis and adds all constraints. */
	// protected Solver getSolver() {
	// Solver solver = translation.createSolver(tracToExpr);
	//
	// if(rawConstraints.size() > 0) {
	// solver.add(rawConstraints.toArray(new BoolExpr[rawConstraints.size()]));
	//
	// Set<BoolExpr> constraintTrackerExprs = constraintTracker2Constraint.keySet();
	// solver.add(translation.getContext().mkOr(
	// constraintTrackerExprs.toArray(new BoolExpr[constraintTrackerExprs.size()])));
	// }
	// return solver;
	// }

	// /** Needs to be overwritten to add the constraints to be checked. */
	// protected void addRawConstraints() {
	// // EMPTY
	// }
	//
	protected abstract void addConstriants();

	protected void translateConstraints() {
		// for(IProductLineConstraint constraint : plConstraints) {
		// Expr[] quantifierValiables = constraint.getQuantifierVariables();
		// EClass[] variableClasses = constraint.getVariablesClasses();
		// BoolExpr[] containmentExprs = new BoolExpr[0];
		// BoolExpr body = constraint.getBody();
		//
		// boolean isForAll = constraint.isForAll();
		//
		// BoolExpr quantifier = createLiftedConstraint(quantifierValiables, variableClasses,
		// containmentExprs, body, isForAll);
		//
		// BoolExpr constraintTracker =
		// ctx.mkBoolConst("CONSTRAINT_" + constraint.getConstraintName());
		// constraintTracker2Constraint.put(constraintTracker, constraint);
		//
		// rawConstraints.add(ctx.mkEq(constraintTracker, quantifier));
		// constraintTrackers.add(constraintTracker);
		// }
	}

	/** Automatically lifts the given constraint */
	protected BoolExpr liftConstraint(Expr[] quantifierValiables, EClass[] variableClasses,
			BoolExpr body, boolean isForAll) {
		return createLiftedConstraint(quantifierValiables, variableClasses, new BoolExpr[0], body,
				isForAll);
	}

	// protected BoolExpr createLiftedQuantifier(String[] quantifierValiableNames,
	// EClass[] variableClasses, BoolExpr[] containmentExprs, BoolExpr body,
	// boolean isForAll) {
	//
	// int numVar = quantifierValiableNames.length;
	// Expr[] quantifierValiables = new Expr[numVar];
	//
	// for(int i = 0; i < numVar; i++) {
	// EnumSort sort = translation.geteClass2Datatype().get(variableClasses[i]);
	// Expr expr = ctx.mkConst(quantifierValiableNames[i], sort);
	//
	// quantifierValiables[i] = expr;
	// }
	//
	// return createLiftedConstraint(quantifierValiables, variableClasses, containmentExprs, body,
	// isForAll);
	// }

	/** Automatically lifts the given constraint */
	protected BoolExpr createLiftedConstraint(Expr[] quantifierValiables, EClass[] variableClasses,
			BoolExpr[] containmentExprs, BoolExpr body, boolean isForAll) {

		BoolExpr[] selExprs = new BoolExpr[quantifierValiables.length + containmentExprs.length];
		for(int i = 0; i < quantifierValiables.length; i++) {
			EClass cls = variableClasses[i];
			Expr var = quantifierValiables[i];

			FuncDecl selFun = translation.getSelectionFunction(cls);
			BoolExpr selected = (BoolExpr)ctx.mkApp(selFun, var);
			BoolExpr notNull = ctx.mkNot(ctx.mkEq(var, translation.getNullElement(cls)));

			selExprs[i] = ctx.mkAnd(selected, notNull);
		}

		// Add containment expressions
		int j = 0;
		for(int i = quantifierValiables.length; i < quantifierValiables.length +
				containmentExprs.length; i++) {
			selExprs[i] = containmentExprs[j++];
		}

		if(isForAll) {
			BoolExpr liftedBody = ctx.mkImplies(ctx.mkAnd(selExprs), body);
			return ctx.mkForall(quantifierValiables, liftedBody, 0, null, null, null, null);
		}

		// else: Exits
		BoolExpr liftedBody = ctx.mkAnd(ctx.mkAnd(selExprs), body);
		return ctx.mkExists(quantifierValiables, liftedBody, 0, null, null, null, null);
	}

	/** Logs an error in the analysis. */
	private void logError(String message) {
		error(getDefault(), "Error in product-line analysis: " + message);
	}
}
